import json as jsonmod
import logging
import typing as t

import aiohttp

from .. import types
from ..models.object import Fetchable, Snowflake, SoteriaObject
from ..models.post import SelfPost
from ..models.user import SelfUser
from .cache import Cache
from .errors import HTTPError

ObjectT = t.TypeVar("ObjectT", bound=SoteriaObject)

_log = logging.getLogger(__name__)

__all__ = ("Client",)


class Client:
    __slots__ = ("_http", "url", "cdn_url", "auth", "cache", "_user")

    _http: t.Optional[aiohttp.ClientSession]
    url: str
    cdn_url: str
    auth: t.Optional[str]
    cache: Cache
    _user: t.Optional[SelfUser]

    def __init__(
        self,
        *,
        auth: t.Optional[str] = None,
        url: t.Optional[str] = None,
        cdn_url: t.Optional[str] = None,
        cache: Cache
    ):
        self._http = None
        self.url = url or "https://api.soteria.social"
        self.cdn_url = cdn_url or (self.url + "/files")
        self.auth = auth
        self.cache = cache

    @property
    def http(self):
        if not self._http:
            self._http = aiohttp.ClientSession()

        return self._http

    # TODO: multipart forms
    async def request(
        self,
        method: str,
        route: str,
        *,
        params: t.Optional[dict[str, t.Any]] = None,
        json: t.Optional[t.Any] = None,
        auth: t.Optional[str] = None,
    ):
        auth = auth or self.auth
        headers = {"Authorization": f"Bearer {auth}"}
        url = self.url + route

        async with self.http.request(
            method,
            url,
            params=params,
            json=json,
            headers=headers,
        ) as resp:
            _log.debug("Made request to %s with method %s.", url, method)

            content = await resp.text()

            if resp.status > 400:
                raise HTTPError(resp.status, content, resp.content_type)

            return jsonmod.loads(content)

    async def close(self):
        if self._http is None:
            return

        await self._http.close()
        self._http = None

    async def get_or_fetch(self, id: Snowflake, type_: type[ObjectT]) -> ObjectT:
        if not issubclass(type_, Fetchable):
            raise TypeError(f"{type_.__name__} cannot be fetched!")

        obj = await self.cache.retrieve(id, type_)

        if obj is not None:
            return obj

        obj = await type_.fetch(self, id)
        return obj

    async def fetch_self(self) -> SelfUser:
        data: types.User = await self.request("GET", "/users/@me")
        user = SelfUser.from_payload(self, data)

        self._user = user
        return user

    @property
    def user(self) -> SelfUser | None:
        return self._user

    async def create_post(self, content: str) -> SelfPost:
        payload: dict[str, t.Any] = {"content": content}
        res: types.Post = await self.request("POST", "/posts", json=payload)

        post = SelfPost.from_payload(self, res)
        await self.cache.store(post)

        return post



import contextlib
import os
import pathlib
import typing as t

import aiohttp
import attrs

if t.TYPE_CHECKING:
    from ..api.client import Client

__all__ = ("Asset",)


def _ensure_path(p: str | os.PathLike[str] | pathlib.Path):
    if not isinstance(p, pathlib.Path):
        p = pathlib.Path(p)

    return p


@attrs.define(slots=True)
class WebReader:
    resp: aiohttp.ClientResponse

    async def __aiter__(self) -> t.AsyncGenerator[t.Any, bytes]:
        stream = self.resp.content
        while not stream.at_eof():
            chunk, _ = await stream.readchunk()
            yield chunk


@attrs.define(kw_only=True, slots=True)
class Asset:
    client: "Client"
    asset_id: str

    @property
    def url(self) -> str:
        return f"{self.client.cdn_url}/{self.asset_id}"
    
    @contextlib.asynccontextmanager
    async def stream(self):
        client = aiohttp.ClientSession()
        req = client.request("GET", self.url)

        try:
            resp = await req.__aenter__()
            yield WebReader(resp)
        finally:
            await req.__aexit__(None, None, None)
            await client.close()

    async def read(self) -> bytes:
        buffer = bytearray()
        async with self.stream() as stream:
            async for chunk in stream:
                buffer.extend(chunk)

        return bytes(buffer)

    async def save(self, path: str | os.PathLike[str] | pathlib.Path) -> None:
        path = _ensure_path(path)
        contents = await self.read()
        path.write_bytes(contents)



from . import object, post, user, asset
from .object import *
from .post import *
from .user import *
from .asset import *

__all__ = ()
__all__ += object.__all__
__all__ += post.__all__
__all__ += user.__all__
__all__ += asset.__all__


import typing as t

__all__ = (
    "User",
    "Followers",
    "Followees",
    "Blocked",
)


class User(t.TypedDict):
    id: str
    username: str
    display_name: t.Optional[str]
    description: t.Optional[str]
    created_at: str
    visible: bool
    avatar: str
    banner: t.Optional[str]


class Followers(t.TypedDict):
    count: int
    followers: list[str]


class Followees(t.TypedDict):
    count: int
    followees: list[str]


class Blocked(t.TypedDict):
    blocked: list[str]



import abc
import typing as t

import attrs
import typing_extensions as te

from .. import types
from .object import Fetchable, Snowflake, SoteriaObject

if t.TYPE_CHECKING:
    from ..api.client import Client
    from .user import SelfUser, User

__all__ = ("Post",)


class BasePost(abc.ABC):
    __slots__ = ()

    client: "Client"

    id: Snowflake
    author_id: Snowflake
    content: str
    # TODO: attachments

    _cached_likes: list["User"]

    @classmethod
    @abc.abstractmethod
    def from_payload(cls: type[te.Self], client: "Client", payload: types.Post) -> te.Self:
        pass

    async def fetch_likes(self) -> list["User"]:
        data: types.Likes = await self.client.request("GET", f"/posts/{self.id}/likes")
        users = [await self.client.get_or_fetch(Snowflake(id), User) for id in data["likes"]]
        self._cached_likes = users

        return users

    @property
    def likes(self) -> list["User"]:
        return self._cached_likes

    async def like(self) -> None:
        await self.client.request("PUT", f"/posts/{self.id}/like")

    async def unlike(self) -> None:
        await self.client.request("DELETE", f"/posts/{self.id}/like")


@attrs.define(kw_only=True, slots=True)
class Post(SoteriaObject, Fetchable, BasePost):
    client: "Client" = attrs.field(repr=False)

    id: Snowflake = attrs.field(repr=True)
    author_id: Snowflake = attrs.field(repr=False)
    content: str = attrs.field(repr=True)

    _cached_author: t.Optional["User"] = attrs.field(init=False, repr=False)
    _cached_likes: list["User"] = attrs.field(init=False, repr=False)

    @classmethod
    def from_payload(cls, client: "Client", payload: types.Post):
        self = cls(
            client=client,
            id=Snowflake(payload["id"]),
            author_id=Snowflake(payload["author"]),
            content=payload["content"],
        )
        return self

    @classmethod
    async def fetch(cls, client: "Client", id: int):
        data: types.Post = await client.request("GET", f"/posts/{id}")
        return cls.from_payload(client, data)

    async def fetch_author(self) -> "User":
        from .user import User

        user = await self.client.get_or_fetch(self.author_id, User)
        self._cached_author = user

        return user

    @property
    def author(self) -> t.Optional["User"]:
        return self._cached_author


@attrs.define(kw_only=True, slots=True)
class SelfPost(SoteriaObject, BasePost):
    client: "Client" = attrs.field(repr=False)

    id: Snowflake = attrs.field(repr=True)
    author_id: Snowflake = attrs.field(repr=False)
    content: str = attrs.field(repr=True)

    _cached_author: t.Optional["SelfUser"] = attrs.field(init=False, repr=False)
    _cached_likes: list["User"] = attrs.field(init=False, repr=False)

    @classmethod
    def from_payload(cls, client: "Client", payload: types.Post):
        self = cls(
            client=client,
            id=Snowflake(payload["id"]),
            author_id=Snowflake(payload["author"]),
            content=payload["content"],
        )
        return self

    async def fetch_author(self) -> "SelfUser":
        self_user = await self.client.fetch_self()
        self._cached_author = self_user

        return self_user

    @property
    def author(self) -> t.Optional["SelfUser"]:
        return self._cached_author

    async def edit(self, content: str):
        payload: dict[str, t.Any] = {"content": content}

        res: types.Post = await self.client.request("PATCH", f"/posts/{id}", json=payload)
        return SelfPost.from_payload(self.client, res)

    async def delete(self):
        await self.client.request("DELETE", f"/posts/{id}")



import typing as t

__all__ = (
    "Post",
    "Likes",
)


class Attachment(t.TypedDict):
    id: str
    type: str
    post: str


class Post(t.TypedDict):
    id: str
    author: str
    content: str
    created_at: str
    attachments: list[Attachment]


class Likes(t.TypedDict):
    count: int
    likes: list[str]



import abc
import typing as t

import attrs
import typing_extensions as te

from .. import types
from .object import Fetchable, Snowflake, SoteriaObject

if t.TYPE_CHECKING:
    from ..api.client import Client
    from .post import Post

__all__ = (
    "User",
    "SelfUser",
)


class BaseUser(abc.ABC):
    __slots__ = ()

    client: "Client"

    id: Snowflake
    username: str
    display_name: t.Optional[str]
    description: t.Optional[str]
    visible: bool
    avatar: str
    banner: t.Optional[str]

    _cached_followers: list["User"]
    _cached_followees: list["User"]
    _cached_posts: list["Post"]
    _cached_liked_posts: list["Post"]

    @classmethod
    @abc.abstractmethod
    def from_payload(cls: type[te.Self], client: "Client", payload: types.User) -> te.Self:
        pass

    async def fetch_followers(self) -> list["User"]:
        data: types.Followers = await self.client.request("GET", f"/users/{self.id}/followers")
        users = [await self.client.get_or_fetch(Snowflake(id), User) for id in data["followers"]]
        self._cached_followers = users

        return users

    async def fetch_followees(self) -> list["User"]:
        data: types.Followees = await self.client.request("GET", f"/users/{self.id}/followees")
        users = [await self.client.get_or_fetch(Snowflake(id), User) for id in data["followees"]]
        self._cached_followees = users

        return users

    async def fetch_posts(self) -> list["Post"]:
        from .post import Post

        data: list[types.Post] = await self.client.request("GET", f"/users/{self.id}/posts")
        posts: list["Post"] = []

        for payload in data:
            if not (obj := await self.client.cache.retrieve(Snowflake(payload["id"]), Post)):
                obj = Post.from_payload(self.client, payload)
                await self.client.cache.store(obj)

            posts.append(obj)

        self._cached_posts = posts
        return posts

    async def fetch_liked_posts(self) -> list["Post"]:
        from .post import Post

        data: list[str] = await self.client.request("GET", f"/users/{self.id}/posts/likes")
        posts: list["Post"] = [await self.client.get_or_fetch(Snowflake(id), Post) for id in data]

        self._cached_liked_posts = posts
        return posts

    @property
    def followers(self) -> list["User"]:
        return self._cached_followers

    @property
    def followees(self) -> list["User"]:
        return self._cached_followees

    @property
    def posts(self) -> list["Post"]:
        return self._cached_posts

    @property
    def liked_posts(self) -> list["Post"]:
        return self._cached_liked_posts


@attrs.define(kw_only=True, slots=True)
class User(SoteriaObject, Fetchable, BaseUser):
    client: "Client" = attrs.field(repr=False)

    id: Snowflake = attrs.field(repr=True)
    username: str = attrs.field(repr=True)
    display_name: t.Optional[str] = attrs.field(repr=True)
    description: t.Optional[str] = attrs.field(repr=False)
    visible: bool = attrs.field(repr=True)
    avatar: str = attrs.field(repr=False)
    banner: t.Optional[str] = attrs.field(repr=False)

    _cached_followers: list["User"] = attrs.field(init=False, repr=False)
    _cached_followees: list["User"] = attrs.field(init=False, repr=False)
    _cached_posts: list["Post"] = attrs.field(init=False, repr=False)
    _cached_liked_posts: list["Post"] = attrs.field(init=False, repr=False)

    @classmethod
    def from_payload(cls, client: "Client", payload: types.User):
        self = cls(
            client=client,
            id=Snowflake(payload["id"]),
            username=payload["username"],
            display_name=payload.get("display_name"),
            description=payload.get("description"),
            visible=payload["visible"],
            avatar=payload["avatar"],
            banner=payload.get("banner"),
        )
        return self

    @classmethod
    async def fetch(cls, client: "Client", id: int):
        data: types.User = await client.request("GET", f"/users/{id}")
        return cls.from_payload(client, data)

    async def follow(self) -> None:
        await self.client.request("POST", f"/users/{self.id}/followers")

    async def unfollow(self) -> None:
        await self.client.request("DELETE", f"/users/{self.id}/followers")


@attrs.define(kw_only=True, slots=True)
class SelfUser(SoteriaObject, BaseUser):
    client: "Client" = attrs.field(repr=False)

    id: Snowflake = attrs.field(repr=True)
    username: str = attrs.field(repr=True)
    display_name: t.Optional[str] = attrs.field(repr=True)
    description: t.Optional[str] = attrs.field(repr=False)
    visible: bool = attrs.field(repr=True)
    avatar: str = attrs.field(repr=False)
    banner: t.Optional[str] = attrs.field(repr=False)

    _cached_followers: list["User"] = attrs.field(init=False, repr=False)
    _cached_followees: list["User"] = attrs.field(init=False, repr=False)
    _cached_blocked: list["User"] = attrs.field(init=False, repr=False)
    _cached_posts: list["Post"] = attrs.field(init=False, repr=False)
    _cached_liked_posts: list["Post"] = attrs.field(init=False, repr=False)

    @classmethod
    def from_payload(cls, client: "Client", payload: types.User):
        self = cls(
            client=client,
            id=Snowflake(payload["id"]),
            username=payload["username"],
            display_name=payload.get("display_name"),
            description=payload.get("description"),
            visible=payload["visible"],
            avatar=payload["avatar"],
            banner=payload.get("banner"),
        )
        return self

    async def fetch_blocked(self) -> list["User"]:
        data: types.Blocked = await self.client.request("GET", "/users/@me/blocked")
        users = [await self.client.get_or_fetch(Snowflake(id), User) for id in data["blocked"]]
        self._cached_blocked = users

        return users

    @property
    def blocked(self) -> list["User"]:
        return self._cached_blocked

    async def edit_account(
        self,
        *,
        username: t.Optional[str] = None,
        email: t.Optional[str] = None,
        password: t.Optional[str] = None,
        visible: t.Optional[bool] = None,
        current_email: t.Optional[str] = None,
        current_password: t.Optional[str] = None,
    ):
        payload: dict[str, t.Any] = dict(
            filter(
                lambda x: x[1] is not None,
                {
                    "username": username,
                    "email": email,
                    "password": password,
                    "visible": visible,
                    "current_email": current_email,
                    "current_password": current_password,
                }.items(),
            )
        )

        res: types.User = await self.client.request("PATCH", "/users/@me", json=payload)
        return SelfUser.from_payload(self.client, res)

    async def edit_profile(
        self,
        *,
        display_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
    ):
        payload: dict[str, t.Any] = dict(
            filter(
                lambda x: x[1] is not None,
                {
                    "display_name": display_name,
                    "description": description,
                }.items(),
            )
        )

        res: types.User = await self.client.request("PATCH", "/users/profile/@me", json=payload)
        return SelfUser.from_payload(self.client, res)



import abc
import datetime
import typing as t

import typing_extensions as te

if t.TYPE_CHECKING:
    from ..api.client import Client

__all__ = (
    "Snowflake",
    "SoteriaObject",
    "Fetchable",
)


class Snowflake(int):
    @property
    def created_at(self):
        return datetime.datetime.fromtimestamp(((self >> 22) - 1717225200) / 1000)


class SoteriaObject:
    __slots__ = ("id",)

    id: Snowflake

    def __init__(self, id: Snowflake):
        self.id = id

    @property
    def created_at(self):
        return self.id.created_at


class Fetchable(abc.ABC):
    @classmethod
    @abc.abstractmethod
    async def fetch(cls: type[te.Self], client: "Client", id: int) -> te.Self:
        pass



import abc
import typing as t

from ..models.object import Snowflake, SoteriaObject

ObjectT = t.TypeVar("ObjectT", bound=SoteriaObject)

__all__ = (
    "Cache",
    "MemoryCache",
)


class Cache(abc.ABC):
    __slots__ = ()

    @abc.abstractmethod
    async def store(self, object: SoteriaObject) -> None:
        pass

    @abc.abstractmethod
    async def retrieve(self, id: Snowflake, type_: type[ObjectT]) -> t.Optional[ObjectT]:
        pass

    @abc.abstractmethod
    async def remove(self, object: SoteriaObject) -> None:
        pass


class MemoryCache(Cache):
    __slots__ = ("_memory",)

    _memory: dict[Snowflake, SoteriaObject]

    def __init__(self) -> None:
        self._memory = {}

    async def store(self, object: SoteriaObject) -> None:
        # TODO: error handling?
        if object.id in self._memory:
            return

        self._memory[object.id] = object

    async def retrieve(self, id: Snowflake, type_: type[ObjectT]) -> t.Optional[ObjectT]:
        obj = self._memory.get(id)

        if not isinstance(obj, type_):
            return None

        return obj

    async def remove(self, object: SoteriaObject) -> None:
        pass



import json
import typing as t

from ..types.errors import HTTPErrorData

__all__ = (
    "SoteriaError",
    "HTTPError",
)


class SoteriaError(Exception):
    pass


class HTTPError(SoteriaError):
    __slots__ = ("status", "data", "message")

    data: t.Union[HTTPErrorData, str]
    message: str
    status: int

    def __init__(self, status: int, content: str, content_type: str):
        self.status: int = status

        if content_type == "application/json":
            self.data = t.cast(HTTPErrorData, json.loads(content))
            self.message = self.data["message"]
        else:
            self.data = content
            self.message = content

        super().__init__(f"Status code ({self.status}), message:\n{self.message}")
